//wssocket.prg
#include "hmg.ch"
#include "hbcompat.ch"
#include "hbwin.ch"
#include "common.ch"

#define _SERVICE_NAME "hbwsocket"
#define _SERVICE_DESCRIBE "Harbour Web Socket"

#define CRLF chr(13)+chr(10)
                
#define MAGIC_KEY "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"


#define PORT_FILE hb_dirBase() + "port.txt"
static oServer, oPP
static nPort, lBackground

extern hb_version

PROCEDURE wssocket( cMode, cPuerto )
   LOCAL nError
   LOCAL cMsg

   PUBLIC oPP := __pp_Init( "c:\harbour\include", "std.ch" )
   
   lBackground = .T.
   
   IF ! File( PORT_FILE ) .OR. ! Empty( cPuerto )
      DEFAULT cPuerto TO "2000"   
      MemoWrit( PORT_FILE, AllTrim( cPuerto ) )
   ENDIF
   //adjust include path
   
   // DEFAULT cMode TO "S" /* NOTE: Must be the default action */

   DEFAULT cMode TO "N" // Run as Application

   SWITCH Upper( cMode )
   CASE "I"

      IF win_serviceInstall( _SERVICE_NAME, _SERVICE_DESCRIBE )
         MsgInfo( "Service has been successfully installed" )
      ELSE
         nError := wapi_GetLastError()
         cMsg := Space( 128 )
         wapi_FormatMessage( ,,,, @cMsg )
         MsgInfo( "Error installing service: " + hb_ntos( nError ) + " " + cMsg )
      ENDIf
      EXIT

   CASE "U"

      IF win_serviceDelete( _SERVICE_NAME )
         MsgInfo( "Service has been deleted" )
         ferase( PORT_FILE )
      ELSE
         nError := wapi_GetLastError()
         cMsg := Space( 128 )
         wapi_FormatMessage( ,,,, @cMsg )
         MsgInfo( "Error deleting service: " + hb_ntos( nError ) + " " + cMsg )
      ENDIf
      EXIT
   
   CASE "N"
      /* NOTE: Used run like a app */      
      lBackground = .F.
      SrvMain()
      EXIT
   CASE "S"
      /* NOTE: Used when starting up as service.
               Do not invoke the executable manually with this option */

      IF win_serviceStart( _SERVICE_NAME, @SrvMain() )
         MsgInfo( "Service has started OK" )
      ELSE
         nError := wapi_GetLastError()
         cMsg := Space( 128 )
         wapi_FormatMessage( ,,,, @cMsg )
         MsgInfo( "Service has had some problems: " + hb_ntos( nError ) + " " + cMsg )
      ENDIF    
   OTHERWISE
      MsgInfo( "Invalid parameter" )
      EXIT

   ENDSWITCH
   RETURN

PROCEDURE SrvMain()
   local oServer
   local cInclude
   local oError
   
   nPort = Val( MemoRead( PORT_FILE ) )
   
   cInclude = hb_getEnv( "INCLUDE" )
   if Empty( cInclude )
      cInclude = hb_getEnv( "INCLUDE_DIR" )
      if Empty( cInclude )
         cInclude = "\harbour\include"
      endif      
   endif
   BEGIN SEQUENCE
   
   oPP = __pp_Init( cInclude, "std.ch" )
   oServer = HBSocket():new( nPort )
      if lBackground
         oServer:bOnProccess = {|| ! ( win_serviceGetStatus() == WIN_SERVICE_RUNNING ) }
      else 
//         ?  "Press <ESC> to exit!!!"
      endif
      oServer:lDebug = .T.
      oServer:bDebug = {| aParam | LogFile( "debug.txt", aParam ) }
      oServer:bOnRead = {| oSrv, oClient | OnReadProcess( oClient ) }
      oServer:Listen()
      oServer:End()
   RECOVER USING oError
      LogFile( "error.log", { oError:description } )
   ENDSEQUENCE

   KillBack()

RETURN

PROCEDURE KillBack()
    win_serviceSetExitCode( 0 )
    win_serviceStop()
RETURN

function OnReadProcess( oClient )
   
   local cMask, cData, n, cMsg:="", oError
   local nMaskPos := 1, cAnswer
   
   //handshake
   if oClient:Cargo == NIL .OR. ( hb_HHasKey( oClient:Cargo, "HANDSHAKE" ) .AND. ! oClient:Cargo[ "HANDSHAKE" ] )
      HandShake( oClient ) 
   else 
      cData = oClient:cBuffer
      cMask = SubStr( cData, 3, 4 )
      for n = 1 to Len( SubStr( cData, 7 ) )
          cMsg += Chr( hb_BitXOR( Asc( SubStr( cMask, nMaskPos++, 1 ) ), Asc( SubStr( cData, 6 + n, 1 ) ) ) )
          if nMaskPos == 5
             nMaskPos = 1
          endif   
      next 
   
      // oClient:SendData( chr( 129 ) + chr( len( cMsg ) ) + hb_StrToUTF8( cMsg ) )
      
      BEGIN SEQUENCE

	cMsg := ALLTRIM(UPPER(cMsg))

	* HERE IS THE IMPORTANT STUFF !!!! *********************************************

	cAnswer := &cMsg

	/*

	IF cMsg == 'PROC_1'

		cAnswer := 'PROC_1 Return Value'

	ELSEIF cMsg == 'PROC_2'

		cAnswer := 'PROC_2 Return Value'

	ELSE

		cAnswer := 'ERROR: Invalid Procedure Name'

	ENDIF

	*/

	* END OF IMPORTANT STUFF !!!! **************************************************

      RECOVER USING oError
         cAnswer = "Error: " + oError:Description
      ENDSEQUENCE      

      oClient:SendData( Chr( 129 ) + Chr( Len( cAnswer ) ) + hb_StrToUtf8( cAnswer ) )
   endif

return nil


function cValToChar ( xValue )
LOCAL cAnswer

      	IF VALTYPE ( xValue ) = 'C'
		cAnswer := xRes
	ELSEIF valtype ( xValue ) = 'N'
		cAnswer := STR( xValue )
	ELSEIF valtype ( xValue ) = 'L'
		cAnswer := IF( xValue ,'.T.','.F.')
	ELSEIF valtype ( xValue ) = 'D'
		cAnswer := DTOC(xValue)
	ELSEIF valtype ( xValue ) = 'A'
		cAnswer := 'ARRAY'
	ELSEIF valtype ( xValue ) = 'O'
		cAnswer := 'OBJECT'
	ELSEIF valtype ( xValue ) = 'U'
		cAnswer := 'UNDEFINED'
	ELSE
		cAnswer := '***ERROR***'
	ENDIF

return (cAnswer)

//-----------------------------------------//

function HandShake(  oClient )
   local nLen, cBuffer, cContext, cKey, cSend
   
   cBuffer := oClient:cBuffer

   cContext = GetContext( cBuffer, "Sec-WebSocket-Key" )
   cKey     = hb_Base64Encode( hb_sha1( cContext + MAGIC_KEY, .T. ) ) // + "." add something to check that the handshake gets wrong
   
   cSend = "HTTP/1.1 101 Switching Protocols" + CRLF + ;
           "Upgrade: websocket" + CRLF + ;
           "Connection: Upgrade" + CRLF + ;
           "Sec-WebSocket-Accept: " + cKey + CRLF + CRLF   

   nLen = oClient:SendData( cSend )

   IF nLen > 0
      oClient:Cargo = hb_Hash()
      oClient:cargo[ "HANDSHAKE" ] = .T.    
      
   ENDIF

   
return nil 

//-----------------------------------------//

function GetContext( cData, cContext )

  local nLen := Len( cContext )
  local cValue := ""
  local aLines
  local aSubLine
  local cRow
  
  aLines := hb_ATokens( cData, CRLF )
  for each cRow in aLines
     if cContext $ cRow
        aSubLine = hb_ATokens( cRow, ":" )
        cValue = AllTrim( aSubLine[ 2 ] )
        exit
     endif
  next

return cValue

//-----------------------------------------//